unit FormGradient;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Palette,Shp_Engine_Image, ExtCtrls, Buttons, StdCtrls, Spin;

type
  TFrmGradient = class(TForm)
    PalettePanel: TPanel;
    cnvPalette: TPaintBox;
    Bevel1: TBevel;
    Button1: TButton;
    Button2: TButton;
    StartColour: TSpinEdit;
    EndColour: TSpinEdit;
    Label1: TLabel;
    Label2: TLabel;
    Panel1: TPanel;
    Panel2: TPanel;
    SpeedButton1: TSpeedButton;
    SpeedButton2: TSpeedButton;
    Label3: TLabel;
    Label4: TLabel;
    ColorDialog: TColorDialog;
    PreviewMe: TCheckBox;
    procedure cnvPalettePaint(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure Button1Click(Sender: TObject);
    procedure StartColourChange(Sender: TObject);
    procedure SpeedButton1Click(Sender: TObject);
    procedure SpeedButton2Click(Sender: TObject);
    procedure EndColourChange(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure cnvPaletteMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure PreviewMeClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
  private
    { Private declarations }
    TempPalette : TPalette;
    procedure UpdatePreview;
  public
    { Public declarations }
  end;

var
  FrmGradient: TFrmGradient;

implementation

uses FormMain;

{$R *.dfm}

procedure TFrmGradient.cnvPalettePaint(Sender: TObject);
var colwidth, rowheight: Real;
    i, j, idx: Integer;
    r: TRect;
begin
     colwidth := cnvPalette.Width / 8;
     rowheight := cnvPalette.Height / 32;
     idx := 0;
     cnvPalette.Canvas.Brush.Color := clbtnface;
     cnvPalette.Canvas.FillRect(rect(0,0,cnvPalette.Width,cnvPalette.Height));
     for i := 0 to 8 do begin
         r.Left := (i * Round(colwidth));

         r.Right := (r.Left + Round(colwidth));
         for j := 0 to 31 do begin
             r.Top := (j * Round(rowheight));
             r.Bottom := (r.Top + Round(rowheight));
             with cnvPalette.Canvas do begin

                  Brush.Color := TempPalette[idx];

                  if (Idx >= StartColour.Value) and (Idx <= StartColour.Value+EndColour.Value) then begin // the current pen
                     Pen.Color := OpositeColour(TempPalette[idx]);
                     //Pen.Mode := pmNotXOR;
                     Rectangle(r.Left,r.Top,r.Right,r.Bottom);
                     MoveTo(r.Left,r.Top);
                     LineTo(r.Right,r.Bottom);
                     MoveTo(r.Right,r.Top);
                     LineTo(r.Left,r.Bottom);
                  end else
                  FillRect(r);
             end;
             Inc(Idx);
         end;
     end;
end;

procedure TFrmGradient.Button2Click(Sender: TObject);
begin
   Close;
end;

procedure TFrmGradient.Button1Click(Sender: TObject);
begin
   PaletteGradient(StartColour.Value,StartColour.Value+EndColour.Value,Panel1.Color,Panel2.Color,SHPPalette);
   FrmMain.cnvpalette.Repaint;
   FrmMain.SetActiveColour(FrmMain.ActiveColour); // Refresh the active Colour, it might have been changed....
   Close;
end;

procedure TFrmGradient.StartColourChange(Sender: TObject);
begin
   {EndColour.MinValue := StartColour.Value+1; // +1 Stops devision by 0 }
   if ((StartColour.Text = '') or (EndColour.Text = '')) then exit;
   EndColour.MaxValue := 255-StartColour.Value;
   EndColour.Value := EndColour.Value;
   // 1.2 Aditions: Preview.
   if PreviewMe.Checked then
   begin
      UpdatePreview;
   end
   else
      cnvPalette.Repaint;
end;

procedure TFrmGradient.SpeedButton1Click(Sender: TObject);
begin
   ColorDialog.Color := Panel1.Color;

   if ColorDialog.Execute then
   begin
      Panel1.Color := ColorDialog.Color;
      // 1.2 Aditions: Preview.
      if PreviewMe.Checked then
      begin
         UpdatePreview;
      end;
   end;
end;

procedure TFrmGradient.SpeedButton2Click(Sender: TObject);
begin
   ColorDialog.Color := Panel2.Color;

   if ColorDialog.Execute then
   begin
      Panel2.Color := ColorDialog.Color;
      // 1.2 Aditions: Preview.
      if PreviewMe.Checked then
      begin
         UpdatePreview;
      end;
   end;
end;

procedure TFrmGradient.EndColourChange(Sender: TObject);
begin
   // 1.2 Aditions: Preview.
   if PreviewMe.Checked then
   begin
      UpdatePreview;
   end
   else
      cnvPalette.Repaint;
end;

procedure TFrmGradient.FormCreate(Sender: TObject);
begin
   PalettePanel.DoubleBuffered := true;
end;

procedure TFrmGradient.cnvPaletteMouseUp(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var colwidth, rowheight: Real;
    i, j, idx: Integer;
begin
     colwidth := cnvPalette.Width / 8;
     rowheight := cnvPalette.Height / 32;
     i := (X div round(colwidth));
     j := (Y div round(rowheight));
     idx := (i * 32) + j;

     // 1.2 code starts here
     //
     // Check if change is valid.
     if idx = StartColour.Value then exit;
     // Set end colour with shift or start colour without it.
     if ssShift in Shift then
     begin
        EndColour.Value := idx - StartColour.Value;
     end
     else
        StartColour.Value := idx;

end;

// 1.2 Adition: Update preview
procedure TFrmGradient.UpdatePreview;
begin
   // Restore TempPalette
   TempPalette := ClonePalette(SHPPalette);
   // Alter palette for preview
   PaletteGradient(StartColour.Value,StartColour.Value+EndColour.Value,Panel1.Color,Panel2.Color,TempPalette);
   // And repaint it.
   cnvPalette.Repaint;
end;

// 1.2: Preview adition
procedure TFrmGradient.PreviewMeClick(Sender: TObject);
begin
   // Restore Original Palette.
   TempPalette := ClonePalette(SHPPalette);
   if PreviewMe.Checked then
   begin
      // Alter palette for preview
      PaletteGradient(StartColour.Value,StartColour.Value+EndColour.Value,Panel1.Color,Panel2.Color,TempPalette);
   end;
   // And repaint it.
   cnvPalette.Repaint;
end;

procedure TFrmGradient.FormShow(Sender: TObject);
begin
   // 1.2 Adition: TempPalette for preview
   PreviewMeClick(Sender);
end;

end.
