unit FormMain;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, Menus, SHP_Engine,SHP_Engine_Image,Palette,
  StdCtrls, Spin, Buttons;

Const
SB_VER = 'v1.21';
SB_TITLE = ' Open Source Palette Editor';
SB_BY = 'Stucuk and Banshee';

type
  TFrmMain = class(TForm)
    MainMenu1: TMainMenu;
    File1: TMenuItem;
    New1: TMenuItem;
    Open1: TMenuItem;
    N1: TMenuItem;
    Exit1: TMenuItem;
    Save1: TMenuItem;
    SaveAs1: TMenuItem;
    N2: TMenuItem;
    Panel1: TPanel;
    Help1: TMenuItem;
    About1: TMenuItem;
    OpenPaletteDialog: TOpenDialog;
    SavePaletteDialog: TSaveDialog;
    ColorDialog: TColorDialog;
    MainPanel: TPanel;
    PalettePanel: TPanel;
    cnvPalette: TPaintBox;
    pnlPalette: TPanel;
    lblActiveColour: TLabel;
    pnlActiveColour: TPanel;
    Panel2: TPanel;
    SpeedButton1: TSpeedButton;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    SpeedButton2: TSpeedButton;
    Red: TSpinEdit;
    Green: TSpinEdit;
    Blue: TSpinEdit;
    Panel4: TPanel;
    Label1: TLabel;
    ools1: TMenuItem;
    Gradient1: TMenuItem;
    CopyColours1: TMenuItem;
    MakeCSchemeFile1: TMenuItem;
    procedure FormCreate(Sender: TObject);
    procedure Exit1Click(Sender: TObject);
    procedure cnvPalettePaint(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure New1Click(Sender: TObject);
    procedure Open1Click(Sender: TObject);
    procedure Save1Click(Sender: TObject);
    procedure SaveAs1Click(Sender: TObject);
    procedure SpeedButton1Click(Sender: TObject);
    procedure RedChange(Sender: TObject);
    procedure SpeedButton2Click(Sender: TObject);
    procedure cnvPaletteMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure About1Click(Sender: TObject);
    procedure Gradient1Click(Sender: TObject);
    procedure CopyColours1Click(Sender: TObject);
    procedure MakeCSchemeFile1Click(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    isEditable : boolean;
    ActiveColour : byte;
    Procedure SetIsEditable(Value:Boolean);
    procedure SetActiveColour(Col: Integer);
  end;

var
  FrmMain: TFrmMain;
  Filename : string; // Internal Filename
  FileFormat : integer;

implementation

uses FormAbout, FormGradient, FormCopyColour,
  FormMakeCScheme;

{$R *.dfm}

procedure TFrmMain.FormCreate(Sender: TObject);
begin
Caption := SB_TITLE + ' ' + SB_VER;
RedChange(Sender);
end;

procedure TFrmMain.Exit1Click(Sender: TObject);
begin
close;
end;

procedure TFrmMain.cnvPalettePaint(Sender: TObject);
var colwidth, rowheight: Real;
    i, j, idx: Integer;
    r: TRect;
begin
     colwidth := cnvPalette.Width / 8;
     rowheight := cnvPalette.Height / 32;
     idx := 0;
     cnvPalette.Canvas.Brush.Color := clbtnface;
     cnvPalette.Canvas.FillRect(rect(0,0,cnvPalette.Width,cnvPalette.Height));
     for i := 0 to 8 do begin
         r.Left := (i * Round(colwidth));

         r.Right := (r.Left + Round(colwidth));
         for j := 0 to 31 do begin
             r.Top := (j * Round(rowheight));
             r.Bottom := (r.Top + Round(rowheight));
             with cnvPalette.Canvas do begin

                  if isEditable then
                  Brush.Color := SHPPalette[idx]
                  else
                  Brush.Color := colourtogray(SHPPalette[idx]);

                  if (Idx = ActiveColour) then begin // the current pen
                     Pen.Color := OpositeColour(SHPPalette[idx]);
                     //Pen.Mode := pmNotXOR;
                     Rectangle(r.Left,r.Top,r.Right,r.Bottom);
                     MoveTo(r.Left,r.Top);
                     LineTo(r.Right,r.Bottom);
                     MoveTo(r.Right,r.Top);
                     LineTo(r.Left,r.Bottom);
                  end else
                  FillRect(r);
             end;
             Inc(Idx);
         end;
     end;
end;

procedure TFrmMain.FormShow(Sender: TObject);
begin
SetIsEditable(False);
CreateBlankPalette;
PalettePanel.DoubleBuffered := true; // Stops Flickering.

New1Click(Sender); // New File

FrmAbout.Label1.Caption := SB_TITLE;
FrmAbout.Label4.Caption := FrmAbout.Label4.Caption + SB_VER;
FrmAbout.Label6.Caption := FrmAbout.Label6.Caption + SB_BY;

FrmAbout.Label2.Caption := FrmAbout.Label2.Caption + SHP_ENGINE_VER;
FrmAbout.Label3.Caption := FrmAbout.Label3.Caption + SHP_ENGINE_BY;
end;

Procedure TFrmMain.SetIsEditable(Value:Boolean);
begin
IsEditable := Value;
cnvPalette.Repaint;
end;

procedure TFrmMain.New1Click(Sender: TObject);
begin
CreateBlankPalette;
SetActiveColour(0);
Caption := SB_TITLE + ' ' + SB_VER + ' - [ Untitled.pal ]';
Filename := ''; // New File, blank filename

SetIsEditable(True);
cnvPalette.Repaint;
end;

procedure TFrmMain.SetActiveColour(Col: Integer);
begin
     ActiveColour := Col;
     if isEditable then
     pnlActiveColour.Color := SHPPalette[ActiveColour]
     else
     pnlActiveColour.Color := colourtogray(SHPPalette[ActiveColour]);
     lblActiveColour.Caption := IntToStr(ActiveColour) + ' (0x' + IntToHex(ActiveColour,3) + ')';

     Red.Value := GetRValue(SHPPalette[ActiveColour]);
     Green.Value := GetGValue(SHPPalette[ActiveColour]);
     Blue.Value := GetBValue(SHPPalette[ActiveColour]);

     cnvPalette.Repaint;
end;

procedure TFrmMain.Open1Click(Sender: TObject);
begin

if OpenPaletteDialog.Execute then
begin

// Select Palette Type To Load
if OpenPaletteDialog.FilterIndex = 1 then
FileFormat := LoadAPaletteFromFile(OpenPaletteDialog.Filename) // Work Out Which Palette To Load
else
begin
if OpenPaletteDialog.FilterIndex = 2 then
LoadPaletteFromFile(OpenPaletteDialog.Filename) // Load TS/RA2 Palette
else
if OpenPaletteDialog.FilterIndex = 3 then
LoadJASCPaletteFromFile(OpenPaletteDialog.Filename); // Load JASC-PAL Palette

FileFormat := OpenPaletteDialog.FilterIndex-1;
end;

cnvPalette.Repaint;
SetActiveColour(0);

Filename := OpenPaletteDialog.Filename;
Caption := SB_TITLE + ' ' + SB_VER + ' - [ '+ExtractFileName(OpenPaletteDialog.Filename)+' ]';

SetIsEditable(True);
end;
end;

procedure TFrmMain.Save1Click(Sender: TObject);
begin
If not isEditable then exit;

if filename = '' then
SaveAs1Click(Sender)
else
if FileFormat = 1 then
SavePaletteToFile(Filename)
else
if FileFormat = 2 then
SavePaletteToJASCFile(Filename);

Caption := SB_TITLE + ' ' + SB_VER + ' - [ '+FileName+' ]';
end;

procedure TFrmMain.SaveAs1Click(Sender: TObject);
begin
If not isEditable then exit;

if SavePaletteDialog.Execute then
begin
FileFormat := SavePaletteDialog.FilterIndex;
Filename := SavePaletteDialog.FileName;
Save1Click(Sender);
end;

end;

procedure TFrmMain.SpeedButton1Click(Sender: TObject);
begin
SHPPalette[ActiveColour] := RGB(Red.Value,Green.Value,Blue.Value);
cnvPalette.Repaint;
end;

procedure TFrmMain.RedChange(Sender: TObject);
begin
if (((Red.Text = '') or (Green.Text = '')) or (Blue.Text = '')) then exit;
Panel4.Color := RGB(Red.Value,Green.Value,Blue.Value);
end;

procedure TFrmMain.SpeedButton2Click(Sender: TObject);
begin
ColorDialog.Color := RGB(Red.Value,Green.Value,Blue.Value);

if ColorDialog.Execute then
begin
Red.Value := GetRValue(ColorDialog.Color);
Green.Value := GetGValue(ColorDialog.Color);
Blue.Value := GetBValue(ColorDialog.Color);
end;
end;

procedure TFrmMain.cnvPaletteMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var colwidth, rowheight: Real;
    i, j, idx: Integer;
begin
If not isEditable then exit;
     colwidth := cnvPalette.Width / 8;
     rowheight := cnvPalette.Height / 32;
     i := (X div round(colwidth));
     j := (Y div round(rowheight));
     idx := (i * 32) + j;

     SetActiveColour(idx);

     cnvPalette.Repaint;
end;

procedure TFrmMain.About1Click(Sender: TObject);
begin
FrmAbout.ShowModal;
end;

procedure TFrmMain.Gradient1Click(Sender: TObject);
begin
FrmGradient.showmodal;
end;

procedure TFrmMain.CopyColours1Click(Sender: TObject);
var
x : integer;
begin

for x := 0 to 255 do
begin
FrmCopyColours.CPalette[x] := SHPPalette[x];
FrmCopyColours.CPalette2[x] := SHPPalette[x];
end;

FrmCopyColours.Label6.Visible := false;
FrmCopyColours.endColour2.Visible := false;

FrmCopyColours.showmodal;

if FrmCopyColours.OKS = true then
for x := FrmCopyColours.StartColour2.Value to FrmCopyColours.StartColour2.Value+FrmCopyColours.EndColour.Value do
SHPPalette[x] := SHPPalette[x-FrmCopyColours.StartColour2.Value+FrmCopyColours.StartColour.Value];

cnvPalette.Repaint;
end;

procedure TFrmMain.MakeCSchemeFile1Click(Sender: TObject);
begin
FrmMakeCScheme.showmodal;
end;

end.
