// PALETTE.PAS
// By Banshee & Stucuk

// Specially modified for OS Palette Editor 1.2, on 04/11/2005
unit Palette;

interface

uses
  Windows, Variants, Graphics,SysUtils, math;

const
  TRANSPARENT = 0;

type
TPalette = array[0..255] of TColor;

var
SHPPalette: TPalette;

procedure LoadPaletteFromFile(Filename : string);
procedure LoadPaletteFromFile2(Filename : string; var Palette : TPalette);
procedure SavePaletteToFile(Filename : string);
procedure GetPaletteFromFile(var Palette:TPalette; Filename : string);
Procedure CreateBlankPalette; // does a gradient effect
Procedure CreateBlankPalette_True; // Makes all 0
procedure LoadJASCPaletteFromFile(Filename : string);
procedure LoadJASCPaletteFromFile2(Filename : string; var Palette:TPalette);
procedure SavePaletteToJASCFile(Filename : string);
procedure PaletteGradient(StartNum,EndNum :byte; StartColour,EndColour : TColor; var OutPut : TPalette);
function ClonePalette(const Input : TPalette): TPalette;
Function LoadAPaletteFromFile(Filename:String) : integer;
Function LoadAPaletteFromFile2(Filename:String; var Palette:TPalette) : integer;

implementation

// Loads TS/RA2 Palette into the SHPPalette
procedure LoadPaletteFromFile(Filename : string);
begin
LoadPaletteFromFile2(filename,SHPPalette);
end;

procedure LoadPaletteFromFile2(Filename : string; var Palette : TPalette);
var
Palette_f : array [0..255] of record
red,green,blue:byte;
end;
    x: Integer;
    F : file;
begin
     // open palette file
     AssignFile(F,Filename);
     Reset(F,1); // file of byte

     BlockRead(F,Palette_f,sizeof(Palette_f));
     CloseFile(F);

     for x := 0 to 255 do
     Palette[x] := RGB(Palette_f[x].red*4,Palette_f[x].green*4,Palette_f[x].blue*4);
end;

// Saves TS/RA2 Palette from the SHPPalette
procedure SavePaletteToFile(Filename : string);
var
Palette_f : array [0..255] of record
red,green,blue:byte;
end;
    x: Integer;
    F : file;
begin

     for x := 0 to 255 do
     begin
     Palette_f[x].red := GetRValue(SHPPalette[x]) div 4;
     Palette_f[x].green := GetGValue(SHPPalette[x]) div 4;
     Palette_f[x].blue := GetBValue(SHPPalette[x]) div 4;
     end;

     AssignFile(F,Filename);
     Rewrite(F,1);

     BlockWrite(F,Palette_f,sizeof(Palette_f));
     CloseFile(F);
end;

procedure GetPaletteFromFile(var Palette:TPalette; Filename : string);
var
Palette_f : array [0..255] of record
red,green,blue:byte;
end;
    x: Integer;
    Colour : string;
    F : file;
begin
     // open palette file
     AssignFile(F,Filename);
     Reset(F,1); // file of byte

     BlockRead(F,Palette_f,sizeof(Palette_f));
     CloseFile(F);

     for x := 0 to 255 do
     begin
        Colour := '$00' + IntToHex(Palette_f[x].blue * 4,2) + IntToHex(Palette_f[x].green * 4,2) + IntToHex(Palette_f[x].red * 4,2);
        Palette[x] := StringToColor(Colour);
     end;

end;

Procedure CreateBlankPalette;
var x : integer;
begin
for x := 0 to 255 do
SHPPalette[x] := RGB(255-x,255-x,255-x);
end;

Procedure CreateBlankPalette_True;
var x : integer;
begin
for x := 0 to 255 do
SHPPalette[x] := 0;
end;

// Loads JASC 8Bit Palette into the SHPPalette
procedure LoadJASCPaletteFromFile(Filename : string);
begin
     LoadJASCPaletteFromFile2(Filename,SHPPalette);
end;

procedure LoadJASCPaletteFromFile2(Filename : string; var Palette:TPalette);
var
Palette_f : array [0..255] of record
red,green,blue:byte;
end;
    signature,binaryvalue:string[10];
    x,colours : Integer;
    F : text;
    R,G,B : byte;
begin

     // open palette file
     AssignFile(F,Filename);
     Reset(F);

     {Jasc format validation}
     readln(F,signature); {JASC-PAL}
     readln(F,binaryvalue); {0100}
     readln(F,colours); {256 (number of colours)}

     if (signature <> 'JASC-PAL') then
     MessageBox(0,'Error: JASC Signature Incorrect','Load Palette Error',0)
     else
     if ((binaryvalue <> '0100') or (colours <> 256)) then
     MessageBox(0,'Error: Palette Must Be 8Bit(256 Colours)','Load Palette Error',0)
     else
     Begin
     {Now, convert colour per colour}
         for x:=0 to 255 do
         begin
            {read source info}
            readln(F,R,G,B);

            {Note: No colour conversion needed since JASC-PAL colours are the same as SHPPalette ones}
            Palette_f[x].red := r;
            Palette_f[x].green := g;
            Palette_f[x].blue := b;

            Palette[x] := RGB(Palette_f[x].red,Palette_f[x].green,Palette_f[x].blue);
         end;
     end;
     CloseFile(F);
end;

// Saves the SHPPalette As JASC-PAL
procedure SavePaletteToJASCFile(Filename : string);
var
    signature,binaryvalue:string[10];
    x,colours : Integer;
    F : text;
begin

     AssignFile(F,Filename);
     Rewrite(F);

     signature := 'JASC-PAL';
     binaryvalue := '0100';
     colours := 256;
     writeln(F,signature); {JASC-PAL}
     writeln(F,binaryvalue); {0100}
     writeln(F,colours); {256 (number of colours)}

     for x := 0 to 255 do
     writeln(F,inttostr(GetRValue(SHPPalette[x]))+' ',inttostr(GetGValue(SHPPalette[x]))+' ',GetBValue(SHPPalette[x]));

     CloseFile(F);
end;

// Altered on OS Palette Editor 1.2 for preview purposes.
procedure PaletteGradient(StartNum,EndNum :byte; StartColour,EndColour : TColor; var OutPut : TPalette);
var
   X,Distance : integer;
   StepR,StepG,StepB : Real;
   R,G,B : integer;
begin
   // Get distance and check its validation.
   Distance := EndNum-StartNum;
   if Distance = 0 then // Catch the Divison By 0's
   begin
      MessageBox(0,'Error: PaletteGradient Needs Start Num And '+#13+#13+'End Num To Be Different Numbers','PaletteGradient Input Error',0);
      Exit;
   end;

   // Now check the size of the steps.
   StepR := (Max(GetRValue(EndColour),GetRValue(StartColour)) - Min(GetRValue(EndColour),GetRValue(StartColour))) / Distance;
   StepG := (Max(GetGValue(EndColour),GetGValue(StartColour)) - Min(GetGValue(EndColour),GetGValue(StartColour))) / Distance;
   StepB := (Max(GetBValue(EndColour),GetBValue(StartColour)) - Min(GetBValue(EndColour),GetBValue(StartColour))) / Distance;

   // Check if the step will be positive or negative
   if GetRValue(EndColour) < GetRValue(StartColour) then
      StepR := -StepR;
   if GetGValue(EndColour) < GetGValue(StartColour) then
      StepG := -StepG;
   if GetBValue(EndColour) < GetBValue(StartColour) then
      StepB := -StepB;

   // Get the starting values
   R := GetRValue(StartColour);
   G := GetGValue(StartColour);
   B := GetBValue(StartColour);

   // 1.21 Fix: Avoid higher step on last colour
   dec(EndNum);
   // Let's start the main loop now:
   for x := StartNum to EndNum do
   begin
      // Avoid that R, G or B get a negative value.
      if Round(R + StepR) < 0 then
         R := 0
      else
         R := Round(R + StepR);

     if Round(G + StepG) < 0 then
        G := 0
     else
        G := Round(G + StepG);

     if Round(B + StepB) < 0 then
        B := 0
     else
        B := Round(B + StepB);

     // Avoid the palette colour from getting over 255 (note: 255 = $ff).
     R := R and $ff;
     G := G and $ff;
     B := B and $ff;

     // Write the value.
     Output[x] := RGB(R,G,B);
   end;
   // 1.21 Fix: Confirm last colour
   inc(EndNum);
   OutPut[EndNum] := EndColour;
end;

// OS Palette Editor 1.2: Copy one palette to another
function ClonePalette(const Input : TPalette): TPalette;
var
   x : byte;
begin
   for x := 0 to 255 do
   begin
      Result[x] := Input[x];
   end;
end;


Function LoadAPaletteFromFile(Filename:String) : integer; // Works out which filetype it is
begin
     Result := LoadAPaletteFromFile2(Filename,SHPPalette);
end;

Function LoadAPaletteFromFile2(Filename:String; var Palette:TPalette) : integer; // Works out which filetype it is
var
    signature:string[10];
    F : text;
begin
     Result := 1; // Assume TS/RA2

     // open palette file
     AssignFile(F,Filename);
     Reset(F);

     {Jasc format validation}
     readln(F,signature); {JASC-PAL}
     CloseFile(F);

     if (signature <> 'JASC-PAL') then // If Signature is not JASC-PAL Assume its a TS/RA2 Palette
     LoadPaletteFromFile2(Filename,Palette)
     else
     begin
     Result := 2;
     LoadJASCPaletteFromFile2(Filename,Palette);
     end;
end;

end.

