unit Voxel_Tools;

{ REVISION HISTORY
  2005/11/23 Fixed ApplyCubedNormals, by adding the Binary3DFloodFill stuff.
  2005/11/10 Added ApplyCubedNormals.
  2004/06/14 Added GetSmoothNormal, used for smoothing single voxel points.
  2004/05/02 Normal Scheme's removed, Apply Normals remade. RemoveRedundantVoxels Remade.
  20020224 Fixed non-current folder bugs in loadnormalschemes,
           added flag NO_FORMS to switch back to ParamStr(0) instead
           of Application.ExeName (for console apps, the extra certainty isn't
           worth 300 KB :-) (since ParamStr(0) works fine on my/most PC's)
  20011218 removed all WriteLns
  20011213 using multi-pass flood filling instead of recursion
  20011213 fixed bug in Internal
}

interface

Uses Voxel,normals,Voxel_Engine,math,math3d;

type TApplyNormalsResult = record
     applied,
{     redundant,     } // Not Needed
     confused: Integer;
end;

type TVoxelSmoothData = record
     Pos : TVector3i;
     V : TVoxelUnpacked;
end;

const
   // 1.2 Cubed Normals Constants:
//   DIST1 = 1; // Not needed
   DIST2 = 0.707106781186547524400844362104849; // sqrt(2)/2
   DIST3 = 0.577350269189625764509148780501957; // sqrt(3)/3


//applies normals
function ApplyNormals(Voxel : TVoxelSection) : TApplyNormalsResult;
function ApplyCubedNormals(Voxel : TVoxelSection) : TApplyNormalsResult;
function RemoveRedundantVoxels(Voxel : TVoxelSection) : integer;
function SmoothNormals(var Vxl : TVoxelSection) : TApplyNormalsResult;
function GetSmoothNormal(var Vxl : TVoxelSection; X,Y,Z,Normal : integer) : integer;

// Used in conjunction with new one to make the ultimate one :D
function RemoveRedundantVoxelsOld(voxel: TVoxelSection): integer;

implementation

{uses SysUtils;}

procedure VecToAngles(const _In: TVector3f ; var angles: TVector3f);
var
  forward_: Single;
  yaw, pitch: Single;
begin
  if (_In.z = 0) and (_In.x = 0) then
  begin
    yaw := 0;
    if (_In.z > 0) then
      pitch := 90
    else
      pitch := 270;
  end
  else
  begin
    if (_In.x <> 0) then
      yaw := Round(ArcTan2(_In.z, _In.x) * 180 / M_PI)
    else if (_In.z > 0) then
      yaw := 90
    else
      yaw := -90;
    if (yaw < 0) then
      yaw := yaw + 360;

    forward_ := sqrt(_In.x*_In.x + _In.z*_In.z);
    pitch := Round(ArcTan2(_In.y, forward_) * 180 / M_PI);
    if (pitch < 0) then
      pitch := pitch + 360;
  end;

  angles.x := -pitch;
  angles.y := yaw;
  angles.z := 0;
end;
       {
Function VectorLength(N : TVector3f) : Single;
begin
Result := sqrt(N.X*N.X + N.Y*N.Y + N.Z*N.Z);
end;   }
            {
Function SubtractVector(V1,V2 : TVector3f) : single;
begin
Result := Max(V1.X,V2.X)-Min(V1.X,V2.X);
Result := Result + Max(V1.Y,V2.Y)-Min(V1.Y,V2.Y);
Result := Result + Max(V1.Z,V2.Z)-Min(V1.Z,V2.Z);
end;

Function SubtractVector3f(V1,V2 : TVector3f) : TVector3f;
begin
Result.X := V1.X-V2.X;
Result.Y := V1.Y-V2.Y;
Result.Z := V1.Z-V2.Z;
end;             }

Function Norm2IndexRA2(N : TVector3f) : Byte;
var
   I : integer;
   thiserr,besterr : single;
   Bestindex : byte;
begin
   Bestindex := 0;
   besterr := VectorDistance(SetVector(0,0,0),SubtractVector(RA2Normals[0],N));
   for i := 1 to 244 do
   begin
      thiserr := VectorDistance(SetVector(0,0,0),SubtractVector(RA2Normals[i],N));
      if thiserr < besterr then
      begin
         besterr:=thiserr;
         bestindex:=i;
      end;
   end;
   Result := Bestindex;
end;

Function Norm2IndexTS(N : TVector3f) : Byte;
var
I : integer;
thiserr,besterr : single;
Bestindex : byte;
begin
   Bestindex := 0;
   besterr := VectorDistance(SetVector(0,0,0),SubtractVector(TSNormals[0],N));
   for i := 1 to 35 do
   begin
      thiserr := VectorDistance(SetVector(0,0,0),SubtractVector(TSNormals[i],N));
      if thiserr < besterr then
      begin
         besterr:=thiserr;
         bestindex:=i;
      end;
   end;
   Result := Bestindex;
end;

Function Norm2Index(N : TVector3f; Vxl : TVoxelSection) : Byte;
begin
   if Vxl.Tailer.Unknown = 2 then
      Result := Norm2IndexTS(N)
   else if Vxl.Tailer.Unknown = 4 then
      Result := Norm2IndexRA2(N);
end;

// 1.2 Adition: Cubed Normalizer
function ApplyCubedNormals(Voxel : TVoxelSection) : TApplyNormalsResult;
const
   LIMZERO = 0.00000001;
type
   TBinaryMap = array of array of array of single; //byte;
   TVector3fMap = array of array of array of TVector3f;

   //*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/
   // Create Binary Map to aid calculating the cubed normals
   //*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/
   procedure CreateBinaryMap(var BinaryMap, FilledMap: TBinaryMap; var FloatMap: TVector3fMap; const Voxel : TVoxelSection);
   var
      x,y,z : integer;
      V : TVoxelUnpacked;
   begin
      // Set memory for binary map. Extra memory avoids bound checking
      SetLength(BinaryMap,Voxel.Tailer.XSize+2,Voxel.Tailer.YSize+2,Voxel.Tailer.ZSize+2);
      SetLength(FilledMap,Voxel.Tailer.XSize+2,Voxel.Tailer.YSize+2,Voxel.Tailer.ZSize+2);
//      SetLength(BinaryMap,(2*Voxel.Tailer.XSize)+2,(2*Voxel.Tailer.YSize)+2,(2*Voxel.Tailer.ZSize)+2);
//      SetLength(FilledMap,(2*Voxel.Tailer.XSize)+2,(2*Voxel.Tailer.YSize)+2,(2*Voxel.Tailer.ZSize)+2);
      SetLength(FloatMap,Voxel.Tailer.XSize+2,Voxel.Tailer.YSize+2,Voxel.Tailer.ZSize+2);

      // Fill Filled Map
      for x := Low(FilledMap) to High(FilledMap) do
         for y := Low(FilledMap[x]) to High(FilledMap[x]) do
            for z := Low(FilledMap[x,y]) to High(FilledMap[x,y]) do
               FilledMap[x,y,z] := 1;

      // Fill Float Map
      for x := Low(FloatMap) to High(FloatMap) do
         for y := Low(FloatMap[x]) to High(FloatMap[x]) do
            for z := Low(FloatMap[x,y]) to High(FloatMap[x,y]) do
            begin
               FloatMap[x,y,z].X := 0;
               FloatMap[x,y,z].Y := 0;
               FloatMap[x,y,z].Z := 0;
            end;

      // All used voxels will receive 1. Unused get 0.
      for x := 0 to Voxel.Tailer.XSize-1 do
         for y := 0 to Voxel.Tailer.YSize-1 do
            for z := 0 to Voxel.Tailer.ZSize-1 do
            begin
               // Get voxel data.
               voxel.GetVoxel(x,y,z,v);
               // Check if it's used.
               if v.Used then
               begin
                  BinaryMap[x+1,y+1,z+1] := 1;
                  FilledMap[x+1,y+1,z+1] := 0;
{
                  BinaryMap[(2*x)+1,(2*y)+1,(2*z)+1] := 1;
                  FilledMap[(2*x)+1,(2*y)+1,(2*z)+1] := 0;
                  BinaryMap[(2*x)+2,(2*y)+1,(2*z)+1] := 1;
                  FilledMap[(2*x)+2,(2*y)+1,(2*z)+1] := 0;
                  BinaryMap[(2*x)+1,(2*y)+2,(2*z)+1] := 1;
                  FilledMap[(2*x)+1,(2*y)+2,(2*z)+1] := 0;
                  BinaryMap[(2*x)+2,(2*y)+2,(2*z)+1] := 1;
                  FilledMap[(2*x)+2,(2*y)+2,(2*z)+1] := 0;
                  BinaryMap[(2*x)+1,(2*y)+1,(2*z)+2] := 1;
                  FilledMap[(2*x)+1,(2*y)+1,(2*z)+2] := 0;
                  BinaryMap[(2*x)+2,(2*y)+1,(2*z)+2] := 1;
                  FilledMap[(2*x)+2,(2*y)+1,(2*z)+2] := 0;
                  BinaryMap[(2*x)+1,(2*y)+2,(2*z)+2] := 1;
                  FilledMap[(2*x)+1,(2*y)+2,(2*z)+2] := 0;
                  BinaryMap[(2*x)+2,(2*y)+2,(2*z)+2] := 1;
                  FilledMap[(2*x)+2,(2*y)+2,(2*z)+2] := 0;
}
               end;
            end;
   end;

   //*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/
   // 3D Binary Flood Fill to restore internal voxels
   //*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/
   procedure Binary3DFloodFill(var FM: TBinaryMap);
   type
      T3DPosition = ^T3DPositionItem;
      T3DPositionItem = record
         x,y,z : integer;
         Next : T3DPosition;
      end;
      // Some basic stuff for the queue above.
      procedure AddPoint (var Start,Final : T3DPosition; x,y,z : integer);
      var
         NewPosition : T3DPosition;
      begin
         // This function adds a point to the queue.
         New(NewPosition);
         NewPosition^.x := x;
         NewPosition^.y := y;
         NewPosition^.z := z;
         NewPosition^.Next := nil;
         if Start <> nil then
         begin
            Final^.Next := NewPosition;
         end
         else
         begin
            Start := NewPosition;
         end;
         Final := NewPosition;
      end;
      // This Gets an Info from a Point and remove it from the queue
      procedure GetPoint (var Start,Final : T3DPosition; var x,y,z : integer);
      var
         Temp : T3DPosition;
      begin // Start will never be nil here. Flood must verify it
         x := Start^.x;
         y := Start^.y;
         z := Start^.z;
         Temp := Start;
         if Final = Start then
            Final := nil;
         Start := Start^.Next;
         Dispose(Temp);
      end;
      // Verify if the point is valid.
      function IsPointOK (const x,y,z,maxx,maxy,maxz : integer) : boolean;
      begin
         result := false;
         if (x < 0) or (x > maxx) then exit;
         if (y < 0) or (y > maxy) then exit;
         if (z < 0) or (z > maxz) then exit;
         result := true;
      end;
   //*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/
   // 3D Binary Flood Fill starts here
   //*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/
   var
      List,Final : T3DPosition;
      x,y,z : integer;
      maxx,maxy,maxz : integer;
   begin
      // Get max values for each.
      maxx := High(FM);
      maxy := High(FM[0]);
      maxz := High(FM[0,0]);

      // Start on (0,0,0);
      List := nil;
      Final := nil;
      AddPoint(List,Final,0,0,0);
      FM[0,0,0] := 0;
      // It flood and fill until the list is over.
      while List <> nil do
      begin
         // Get the currently position
         GetPoint(List,Final,x,y,z);
         // Check and add neighboors (6 faces)
         if IsPointOK(x-1,y,z,maxx,maxy,maxz) then
            if FM[x-1,y,z] = 1 then
            begin
               FM[x-1,y,z] := 0;
               AddPoint(List,Final,x-1,y,z);
            end;
         if IsPointOK(x+1,y,z,maxx,maxy,maxz) then
            if FM[x+1,y,z] = 1 then
            begin
               FM[x+1,y,z] := 0;
               AddPoint(List,Final,x+1,y,z);
            end;
         if IsPointOK(x,y-1,z,maxx,maxy,maxz) then
            if FM[x,y-1,z] = 1 then
            begin
               FM[x,y-1,z] := 0;
               AddPoint(List,Final,x,y-1,z);
            end;
         if IsPointOK(x,y+1,z,maxx,maxy,maxz) then
            if FM[x,y+1,z] = 1 then
            begin
               FM[x,y+1,z] := 0;
               AddPoint(List,Final,x,y+1,z);
            end;
         if IsPointOK(x,y,z-1,maxx,maxy,maxz) then
            if FM[x,y,z-1] = 1 then
            begin
               FM[x,y,z-1] := 0;
               AddPoint(List,Final,x,y,z-1);
            end;
         if IsPointOK(x,y,z+1,maxx,maxy,maxz) then
            if FM[x,y,z+1] = 1 then
            begin
               FM[x,y,z+1] := 0;
               AddPoint(List,Final,x,y,z+1);
            end;
      end;
   end;
   //*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/
   // Merge Filled Map on Binary Map
   //*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/
   procedure MergeBinaryMaps(const Source : TBinaryMap; var Destiny : TBinaryMap);
   var
      x,y,z : integer;
   begin
      // Set '1' on destination for every '1' on source.
      for x := 0 to High(Source) do
         for y := 0 to High(Source[x]) do
            for z := 0 to High(Source[x,y]) do
            begin
               if Source[x,y,z] = 1 then
                  Destiny[x,y,z] := 1;
            end;
   end;

   //*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/
   // Build the influence area.
   //*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/

   procedure BuildInfluenceArea(var Map : TBinaryMap);
   var
      x,y,z : integer;
      StartPoint,FinalPoint : integer;
   begin
      // Scan on z direction.
      for x := Low(Map) to High(Map) do
         for y := Low(Map) to High(Map[x]) do
         begin
            // Lower scan
            z := Low(Map[x,y]);
            StartPoint := -1;
            while (z <= High(Map[x,y])) and (StartPoint = -1) do
            begin
               if Map[x,y,z] = 1 then
               begin
                  StartPoint := z;
               end;
               inc(z);
            end;
            // Higher scan (only happens if the lower one worked)
            if StartPoint <> -1 then
            begin
               z := High(Map[x,y]);
               FinalPoint := -1;
               while (z >= Low(Map[x,y])) and (FinalPoint = -1) do
               begin
                  if Map[x,y,z] = 1 then
                  begin
                     FinalPoint := z;
                  end;
                  dec(z);
               end;
               // Now, we do the painting.
               z := StartPoint;
               while z < FinalPoint do
               begin
                  if Map[x,y,z] = 0 then
                  begin
                     Map[x,y,z] := LIMZERO;
                  end;
                  inc(z);
               end;
            end;
         end;

      // Scan on x direction.
      for y := Low(Map[0]) to High(Map[0]) do
         for z := Low(Map[0,y]) to High(Map[0,y]) do
         begin
            // Lower scan
            x := Low(Map);
            StartPoint := -1;
            while (x <= High(Map)) and (StartPoint = -1) do
            begin
               if Map[x,y,z] = 1 then
               begin
                  StartPoint := x;
               end;
               inc(x);
            end;
            // Higher scan (only happens if the lower one worked)
            if StartPoint <> -1 then
            begin
               x := High(Map);
               FinalPoint := -1;
               while (x >= Low(Map)) and (FinalPoint = -1) do
               begin
                  if Map[x,y,z] = 1 then
                  begin
                     FinalPoint := x;
                  end;
                  dec(x);
               end;
               // Now, we do the painting.
               x := StartPoint;
               while x < FinalPoint do
               begin
                  if Map[x,y,z] = 0 then
                  begin
                     Map[x,y,z] := LIMZERO;
                  end;
                  inc(x);
               end;
            end;
         end;

      // Scan on y direction.
      for x := Low(Map) to High(Map) do
         for z := Low(Map[x,0]) to High(Map[x,0]) do
         begin
            // Lower scan
            y := Low(Map[x]);
            StartPoint := -1;
            while (y <= High(Map[x])) and (StartPoint = -1) do
            begin
               if Map[x,y,z] = 1 then
               begin
                  StartPoint := y;
               end;
               inc(y);
            end;
            // Higher scan (only happens if the lower one worked)
            if StartPoint <> -1 then
            begin
               y := High(Map[x]);
               FinalPoint := -1;
               while (y >= Low(Map[x])) and (FinalPoint = -1) do
               begin
                  if Map[x,y,z] = 1 then
                  begin
                     FinalPoint := y;
                  end;
                  dec(y);
               end;
               // Now, we do the painting.
               y := StartPoint;
               while y < FinalPoint do
               begin
                  if Map[x,y,z] = 0 then
                  begin
                     Map[x,y,z] := LIMZERO;
                  end;
                  inc(y);
               end;
            end;
         end;
   end;

   //*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/
   // Calculate new normals
   //*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/
   procedure GetPreliminaryNormals(const BM: TBinaryMap; var FloatMap : TVector3fMap; var V : TVoxelUnpacked; _x,_y,_z : integer);
   var
      res : TVector3f;
      x,y,z : integer;
      xx,yy,zz : byte;
   begin
      // Security checkup.
      if not V.Used then exit;

      res.X := 0;
      res.Y := 0;
      res.Z := 0;


      x := 1 + _x;
      y := 1 + _y;
      z := 1 + _z;

      // Time to get a rx, ry and rz from the binary map
      // Attention: Big lines detected!
      res.X := (BM[x-1,y+1,z+1] * DIST3) + (BM[x-1,y,z+1] * DIST2) + (BM[x-1,y-1,z+1] * DIST3) + (BM[x-1,y+1,z] * DIST2) + BM[x-1,y,z] + (BM[x-1,y-1,z] * DIST2) + (BM[x-1,y+1,z-1] * DIST3) + (BM[x-1,y,z-1] * DIST2) + (BM[x-1,y-1,z-1] * DIST3) - (BM[x+1,y+1,z+1] * DIST3) - (BM[x+1,y,z+1] * DIST2) - (BM[x+1,y-1,z+1] * DIST3) - (BM[x+1,y+1,z] * DIST2) - BM[x+1,y,z] - (BM[x+1,y-1,z] * DIST2) - (BM[x+1,y+1,z-1] * DIST3) - (BM[x+1,y,z-1] * DIST2) - (BM[x+1,y-1,z-1] * DIST3);
      res.Y := (BM[x+1,y-1,z+1] * DIST3) + (BM[x,y-1,z+1] * DIST2) + (BM[x-1,y-1,z+1] * DIST3) + (BM[x+1,y-1,z] * DIST2) + BM[x,y-1,z] + (BM[x-1,y-1,z] * DIST2) + (BM[x+1,y-1,z-1] * DIST3) + (BM[x,y-1,z-1] * DIST2) + (BM[x-1,y-1,z-1] * DIST3) - (BM[x+1,y+1,z+1] * DIST3) - (BM[x,y+1,z+1] * DIST2) - (BM[x-1,y+1,z+1] * DIST3) - (BM[x+1,y+1,z] * DIST2) - BM[x,y+1,z] - (BM[x-1,y+1,z] * DIST2) - (BM[x+1,y+1,z-1] * DIST3) - (BM[x,y+1,z-1] * DIST2) - (BM[x-1,y+1,z-1] * DIST3);
      res.Z := (BM[x+1,y+1,z-1] * DIST3) + (BM[x+1,y,z-1] * DIST2) + (BM[x+1,y-1,z-1] * DIST3) + (BM[x,y+1,z-1] * DIST2) + BM[x,y,z-1] + (BM[x,y-1,z-1] * DIST2) + (BM[x-1,y+1,z-1] * DIST3) + (BM[x-1,y,z-1] * DIST2) + (BM[x-1,y-1,z-1] * DIST3) - (BM[x-1,y+1,z+1] * DIST3) - (BM[x-1,y,z+1] * DIST2) - (BM[x-1,y-1,z+1] * DIST3) - (BM[x,y+1,z+1] * DIST2) - BM[x,y,z+1] - (BM[x,y-1,z+1] * DIST2) - (BM[x+1,y+1,z+1] * DIST3) - (BM[x+1,y,z+1] * DIST2) - (BM[x+1,y-1,z+1] * DIST3);
//    res.X := (BM[x+1,y+1,z+1] * DIST3) + (BM[x+1,y,z+1] * DIST2) + (BM[x+1,y-1,z+1] * DIST3) + (BM[x+1,y+1,z] * DIST2) + BM[x+1,y,z] + (BM[x+1,y-1,z] * DIST2) + (BM[x+1,y+1,z-1] * DIST3) + (BM[x+1,y,z-1] * DIST2) + (BM[x+1,y-1,z-1] * DIST3) - (BM[x-1,y+1,z+1] * DIST3) - (BM[x-1,y,z+1] * DIST2) - (BM[x-1,y-1,z+1] * DIST3) - (BM[x-1,y+1,z] * DIST2) - BM[x-1,y,z] - (BM[x-1,y-1,z] * DIST2) - (BM[x-1,y+1,z-1] * DIST3) - (BM[x-1,y,z-1] * DIST2) - (BM[x-1,y-1,z-1] * DIST3);
//    res.Y := (BM[x+1,y+1,z+1] * DIST3) + (BM[x,y+1,z+1] * DIST2) + (BM[x-1,y+1,z+1] * DIST3) + (BM[x+1,y+1,z] * DIST2) + BM[x,y+1,z] + (BM[x-1,y+1,z] * DIST2) + (BM[x+1,y+1,z-1] * DIST3) + (BM[x,y+1,z-1] * DIST2) + (BM[x-1,y+1,z-1] * DIST3) - (BM[x+1,y-1,z+1] * DIST3) - (BM[x,y-1,z+1] * DIST2) - (BM[x-1,y-1,z+1] * DIST3) - (BM[x+1,y-1,z] * DIST2) - BM[x,y-1,z] - (BM[x-1,y-1,z] * DIST2) - (BM[x+1,y-1,z-1] * DIST3) - (BM[x,y-1,z-1] * DIST2) - (BM[x-1,y-1,z-1] * DIST3);
//    res.Z := (BM[x+1,y+1,z+1] * DIST3) + (BM[x+1,y,z+1] * DIST2) + (BM[x+1,y-1,z+1] * DIST3) + (BM[x,y+1,z+1] * DIST2) + BM[x,y,z+1] + (BM[x,y-1,z+1] * DIST2) + (BM[x-1,y+1,z+1] * DIST3) + (BM[x-1,y,z+1] * DIST2) + (BM[x-1,y-1,z+1] * DIST3) - (BM[x-1,y+1,z-1] * DIST3) - (BM[x-1,y,z-1] * DIST2) - (BM[x-1,y-1,z-1] * DIST3) - (BM[x,y+1,z-1] * DIST2) - BM[x,y,z-1] - (BM[x,y-1,z-1] * DIST2) - (BM[x+1,y+1,z-1] * DIST3) - (BM[x+1,y,z-1] * DIST2) - (BM[x+1,y-1,z-1] * DIST3);

      FloatMap[_x+1,_y+1,_z+1].X := res.X;
      FloatMap[_x+1,_y+1,_z+1].Y := res.Y;
      FloatMap[_x+1,_y+1,_z+1].Z := res.Z;

{
      for xx := 0 to 1 do
      begin
         x := 1 + (2 * _x) + xx;
         for yy := 0 to 1 do
         begin
            y := 1 + (2 * _y) + yy;
            for zz := 0 to 1 do
            begin
               z := 1 + (2 * _z) + zz;

               // Time to get a rx, ry and rz from the binary map
               // Attention: Big lines detected!
               res.X := res.X + (BM[x-1,y+1,z+1] * DIST3) + (BM[x-1,y,z+1] * DIST2) + (BM[x-1,y-1,z+1] * DIST3) + (BM[x-1,y+1,z] * DIST2) + BM[x-1,y,z] + (BM[x-1,y-1,z] * DIST2) + (BM[x-1,y+1,z-1] * DIST3) + (BM[x-1,y,z-1] * DIST2) + (BM[x-1,y-1,z-1] * DIST3) - (BM[x+1,y+1,z+1] * DIST3) - (BM[x+1,y,z+1] * DIST2) - (BM[x+1,y-1,z+1] * DIST3) - (BM[x+1,y+1,z] * DIST2) - BM[x+1,y,z] - (BM[x+1,y-1,z] * DIST2) - (BM[x+1,y+1,z-1] * DIST3) - (BM[x+1,y,z-1] * DIST2) - (BM[x+1,y-1,z-1] * DIST3);
               res.Y := res.Y + (BM[x+1,y-1,z+1] * DIST3) + (BM[x,y-1,z+1] * DIST2) + (BM[x-1,y-1,z+1] * DIST3) + (BM[x+1,y-1,z] * DIST2) + BM[x,y-1,z] + (BM[x-1,y-1,z] * DIST2) + (BM[x+1,y-1,z-1] * DIST3) + (BM[x,y-1,z-1] * DIST2) + (BM[x-1,y-1,z-1] * DIST3) - (BM[x+1,y+1,z+1] * DIST3) - (BM[x,y+1,z+1] * DIST2) - (BM[x-1,y+1,z+1] * DIST3) - (BM[x+1,y+1,z] * DIST2) - BM[x,y+1,z] - (BM[x-1,y+1,z] * DIST2) - (BM[x+1,y+1,z-1] * DIST3) - (BM[x,y+1,z-1] * DIST2) - (BM[x-1,y+1,z-1] * DIST3);
               res.Z := res.Z + (BM[x+1,y+1,z-1] * DIST3) + (BM[x+1,y,z-1] * DIST2) + (BM[x+1,y-1,z-1] * DIST3) + (BM[x,y+1,z-1] * DIST2) + BM[x,y,z-1] + (BM[x,y-1,z-1] * DIST2) + (BM[x-1,y+1,z-1] * DIST3) + (BM[x-1,y,z-1] * DIST2) + (BM[x-1,y-1,z-1] * DIST3) - (BM[x-1,y+1,z+1] * DIST3) - (BM[x-1,y,z+1] * DIST2) - (BM[x-1,y-1,z+1] * DIST3) - (BM[x,y+1,z+1] * DIST2) - BM[x,y,z+1] - (BM[x,y-1,z+1] * DIST2) - (BM[x+1,y+1,z+1] * DIST3) - (BM[x+1,y,z+1] * DIST2) - (BM[x+1,y-1,z+1] * DIST3);
            end;
         end;
      end;
      FloatMap[_x+1,_y+1,_z+1].X := res.X / 8;
      FloatMap[_x+1,_y+1,_z+1].Y := res.Y / 8;
      FloatMap[_x+1,_y+1,_z+1].Z := res.Z / 8;
}
   end;


   procedure GetNewNormals(const BM: TVector3fMap; var V : TVoxelUnpacked; x,y,z : integer);
      function GetNonZeroSign(var value : single) : shortint;
      begin
         Result := sign(value);
         if Result = 0 then
            Result := 1;
      end;
      function GetTrueSign(var value,signal : single; minVar, maxVar, Direction : integer) : shortint;
      var
         V : TVoxelUnpacked;
      begin
         Result := sign(signal);
         if Result = 0 then
         begin
            Result := sign(value);
            if Result = 0 then
            begin
               // we need to do some magic here.
               if minVar <= (maxVar div 2) then
                  Result := -1 * Direction
               else
                  Result := 1 * Direction;
            end;
         end;
      end;
   var
      res : TVector3f;
      signal : TVector3f;
      newvalue : single;
      xx,yy,zz : integer;
      xcounter,ycounter,zcounter : single;
      // 1.3 Upgrade
      Dist : array [0..3] of single;
      Alpha : byte;
      CurrentVoxel : TVoxelUnpacked;
   begin
      // Security checkup.
      if not V.Used then exit;

      res.X := 0;
      res.Y := 0;
      res.Z := 0;
      signal.X := 0;
      signal.Y := 0;
      signal.Z := 0;

      xcounter := 0;
      ycounter := 0;
      zcounter := 0;

      // 1.31 Upgrade
      Dist[0] := 1;
      Dist[1] := 1;
      Dist[2] := DIST2;
      Dist[3] := DIST3;

      // Let's grab the new res
      for xx := x-1 to x+1 do
         for yy := y-1 to y+1 do
            for zz := z-1 to z+1 do
            begin
               Alpha := abs(x - xx) + abs(y - yy) + abs(z - zz);
               if Voxel.GetVoxelSafe(xx-1,yy-1,zz-1,CurrentVoxel) then
                  if CurrentVoxel.Used then
                  begin
                     newvalue := Dist[Alpha] * BM[xx,yy,zz].X;
                     res.X := res.X + (NewValue * GetNonZeroSign(BM[xx,yy,zz].X));
                     signal.X := signal.X + NewValue;
                     xcounter := xcounter + Dist[Alpha];
                     NewValue := Dist[Alpha] * BM[xx,yy,zz].Y;
                     res.Y := res.Y + (NewValue * GetNonZeroSign(BM[xx,yy,zz].Y));
                     signal.X := signal.X + NewValue;
                     ycounter := ycounter + Dist[Alpha];
                     NewValue := Dist[Alpha] * BM[xx,yy,zz].Z;
                     res.Z := res.Z + (NewValue * GetNonZeroSign(BM[xx,yy,zz].Z));
                     signal.Z := signal.Z + NewValue;
                     zcounter := zcounter + Dist[Alpha];
                  end;
            end;
      if xcounter > 0 then
         res.X := (res.X / xcounter) * GetTrueSign(BM[x,y,z].X,signal.X,x-1,Voxel.Tailer.XSize,-1);
      if ycounter > 0 then
         res.Y := (res.Y / ycounter) * GetTrueSign(BM[x,y,z].Y,signal.Y,y-1,Voxel.Tailer.YSize,-1);
      if zcounter > 0 then
         res.Z := (res.Z / zcounter) * GetTrueSign(BM[x,y,z].Z,signal.Z,z-1,Voxel.Tailer.ZSize,1);

      // Normalization process.
      Normalize(res);
      if Voxel.Tailer.Unknown = 4 then
         V.Normal := Norm2IndexRA2(res)
      else
         V.Normal := Norm2IndexTS(res);

      // Apply
      Voxel.SetVoxel(x-1,y-1,z-1,V);
      inc(Result.applied);
   end;
//*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/
// Cubed Normals 2x Main Funcion Starts Here
//*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/
var
   BinaryMap,FilledMap : TBinaryMap;
   FloatMap : TVector3fMap;
   x,y,z : integer;
   V : TVoxelUnpacked;
begin
   Result.applied := 0;
   // Create the binary map
   CreateBinaryMap(BinaryMap,FilledMap,FloatMap,Voxel);
   // 1.2c: Solves inaucurate cubed normalizer
   // This will fill the internal part for acurate results
   Binary3DFloodFill(FilledMap);
   MergeBinaryMaps(FilledMap,BinaryMap);
   // 1.32: Solves limit (x,y,z) -> (0,0,0) on cubed normalizer.
   BuildInfluenceArea(BinaryMap);
   // Now, let's normalize every voxel.
   for x := 0 to Voxel.Tailer.XSize-1 do
      for y := 0 to Voxel.Tailer.YSize-1 do
         for z := 0 to Voxel.Tailer.ZSize-1 do
         begin
            // Get voxel data and calculate it (added +1 to
            // each, since binary map has propositally a
            // border to avoid bound checking).
            Voxel.GetVoxel(x,y,z,v);
            GetPreliminaryNormals(BinaryMap,FloatMap,V,x,y,z);
         end;
   for x := 0 to Voxel.Tailer.XSize-1 do
      for y := 0 to Voxel.Tailer.YSize-1 do
         for z := 0 to Voxel.Tailer.ZSize-1 do
         begin
            Voxel.GetVoxel(x,y,z,v);
            GetNewNormals(FloatMap,V,x+1,y+1,z+1);
         end;

   // Now, let's free some memory.
   for x := Low(BinaryMap) to High(BinaryMap) do
   begin
      for y := Low(BinaryMap[x]) to High(BinaryMap[x]) do
      begin
         SetLength(BinaryMap[x,y],0);
         SetLength(FilledMap[x,y],0);
      end;
      SetLength(BinaryMap[x],0);
      SetLength(FilledMap[x],0);
   end;
   for x := Low(FloatMap) to High(FloatMap) do
   begin
      for y := Low(FloatMap[x]) to High(FloatMap[x]) do
      begin
         SetLength(FloatMap[x,y],0);
      end;
      SetLength(FloatMap[x],0);
   end;

   SetLength(BinaryMap,0);
   SetLength(FilledMap,0);
   SetLength(FloatMap,0);
end;

function ApplyNormals(Voxel : TVoxelSection) : TApplyNormalsResult;
var maxx, maxy, maxz,
    x, y, z: integer;
function Empty(x,y,z: integer): boolean;
var v: TVoxelUnpacked;
begin
     // check bounds
     Result := True; // outside the voxel is all empty
     if (x < 0) or (x > maxx) then Exit;
     if (y < 0) or (y > maxy) then Exit;
     if (z < 0) or (z > maxz) then Exit;
     // ok, do the real check
     voxel.GetVoxel(x,y,z,v);
     if v.Used then
        Result := False
     else
         Result := true;//BitSet(v.Flags,isOutside);
end;
function CheckTopBottom(var b1,b2 : boolean) : boolean;
var
ZZ : integer;
First,Second : boolean;
v : TVoxelunpacked;
begin
   First := True;
   if z < maxz then
     for ZZ := z+1 to Maxz do
        if First then
        begin
           voxel.GetVoxel(x,y,ZZ,v);
           if v.Used then
              First := false;
        end;

   if z = maxz then
      First := true;

   Second := True;

   if z > 0 then
      for ZZ := z-1 downto 0 do
         if Second then
         begin
            voxel.GetVoxel(x,y,ZZ,v);
            if v.Used then
               Second := false;
         end;

   if z = 0 then
      Second := true;

   Result := false;

   If (First) and (not Second) then
   begin
      B2 := true;
      B1 := false;
      Result := true;
   end
   else If (not First) and (Second) then
   begin
      B2 := false;
      B1 := true;
      Result := true;
   end;

end;

function CheckRightLeft(var b1,b2 : boolean) : boolean;
var
YY : integer;
First,Second : boolean;
v : TVoxelunpacked;
begin

First := True;

if y < maxy then
for YY := y+1 to Maxy do
if First then
begin
voxel.GetVoxel(x,yy,z,v);
if v.Used then
First := false;
end;

if y = maxy then
First := true;

Second := True;

if y > 0 then
for YY := y-1 downto 0 do
if Second then
begin
voxel.GetVoxel(x,yy,z,v);
if v.Used then
Second := false;
end;

if y = 0 then
Second := true;

Result := false;

If (First) and (not Second) then
begin
B2 := true;
B1 := false;
Result := true;
end
else
If (not First) and (Second) then
begin
B2 := false;
B1 := true;
Result := true;
end;

end;

function CheckFrontBack(var b1,b2 : boolean) : boolean;
var
XX : integer;
First,Second : boolean;
v : TVoxelunpacked;
begin

First := True;

if x < maxx then
for XX := x+1 to Maxx do
if First then
begin
voxel.GetVoxel(xx,y,z,v);
if v.Used then
First := false;
end;

if x = maxx then
First := true;

Second := True;

if x > 0 then
for xx := x-1 downto 0 do
if Second then
begin
voxel.GetVoxel(xx,y,z,v);
if v.Used then
Second := false;
end;

if x = 0 then
Second := true;

Result := false;

If (First) and (not Second) then
begin
B2 := false;
B1 := true;
Result := true;
end
else
If (not First) and (Second) then
begin
B2 := true;
B1 := false;
Result := true;
end;

end;
Procedure CalcFacing;
var v: TVoxelUnpacked;
    B : array [0..5] of boolean;
    I : integer;
    Pos : TVector3i;
begin
     // get the voxel in question
     voxel.GetVoxel(x,y,z,v);
     // skip empty ones
     if not v.Used then Exit;

     for i := 0 to 5 do
     B[i] := false;

     if Empty(x+1,y,z) then B[0] := true;
     if Empty(x-1,y,z) then B[1] := true;
     if Empty(x,y-1,z) then B[2] := true;
     if Empty(x,y+1,z) then B[3] := true;
     if Empty(x,y,z-1) then B[4] := true;
     if Empty(x,y,z+1) then B[5] := true;

     // cancel out double-sided facings
     if B[0] and B[1] then if not CheckFrontBack(B[0],B[1]) then if x > trunc(maxx/2) then B[1] := false else B[0] := false;  //Dec(Flags,ChooseFrontOrBack(x,y,z));
     if B[2] and B[3] then if not CheckRightLeft(B[2],B[3]) then if y < trunc(maxy/2) then B[3] := false else B[2] := false;
     if B[4] and B[5] then if not CheckTopBottom(B[4],B[5]) then if z < trunc(maxz/2) then B[5] := false else B[4] := false;

     Pos := SetVectorI(0,0,0);
     if B[0] then
     Pos.X := Pos.X + 1;
     if B[1] then
     Pos.X := Pos.X - 1;
     if B[2] then
     Pos.Y := Pos.Y - 1;
     if B[3] then
     Pos.Y := Pos.Y + 1;
     if B[4] then
     Pos.Z := Pos.Z - 1;
     if B[5] then
     Pos.Z := Pos.Z + 1;

     if (Pos.x = 0) and (Pos.z = 0) and (Pos.y = 0) then
     inc(Result.confused)
     else
     begin
     if Voxel.Tailer.Unknown = 4 then
     v.Normal := Norm2IndexRA2(SetVector(Pos.X,Pos.Y,Pos.Z))
     else
     v.Normal := Norm2IndexTS(SetVector(Pos.X,Pos.Y,Pos.Z));

     inc(Result.applied);

     voxel.SetVoxel(x,y,z,v);
     end;
end;
begin
     maxx := voxel.Tailer.XSize - 1;
     maxy := voxel.Tailer.YSize - 1;
     maxz := voxel.Tailer.ZSize - 1;

     Result.applied := 0;
     Result.confused := 0;
     //Result.redundant := 0; // Not used!

     // examine each used voxel, pass one for outside voxels
     for x := 0 to maxx do
         for y := 0 to maxy do
             for z := 0 to maxz do
                 CalcFacing;
end;

function RemoveRedundantVoxels(voxel: TVoxelSection): integer;
var maxx, maxy, maxz,
    x, y, z: integer;
Red : Array of TVector3i;
Red_No : integer;
v: TVoxelUnpacked;
function Empty(x,y,z: integer): boolean;
var v: TVoxelUnpacked;
begin
     // check bounds
     Result := True; // outside the voxel is all empty
     if (x < 0) or (x > maxx) then Exit;
     if (y < 0) or (y > maxy) then Exit;
     if (z < 0) or (z > maxz) then Exit;
     // ok, do the real check
     voxel.GetVoxel(x,y,z,v);
     if v.Used then
        Result := False
     else
         Result := true;//BitSet(v.Flags,isOutside);
end;
function CheckTopBottom(var b1,b2 : boolean) : boolean;
var
ZZ : integer;
First,Second : boolean;
v : TVoxelunpacked;
begin

First := True;

if z < maxz then
for ZZ := z+1 to Maxz do
if First then
begin
voxel.GetVoxel(x,y,ZZ,v);
if v.Used then
First := false;
end;

if z = maxz then
First := true;

Second := True;

if z > 0 then
for ZZ := z-1 downto 0 do
if Second then
begin
voxel.GetVoxel(x,y,ZZ,v);
if v.Used then
Second := false;
end;

if z = 0 then
Second := true;

Result := false;

If (First) and (not Second) then
begin
B2 := true;
B1 := false;
Result := true;
end
else
If (not First) and (Second) then
begin
B2 := false;
B1 := true;
Result := true;
end;

end;

function CheckRightLeft(var b1,b2 : boolean) : boolean;
var
YY : integer;
First,Second : boolean;
v : TVoxelunpacked;
begin

First := True;

if y < maxy then
for YY := y+1 to Maxy do
if First then
begin
voxel.GetVoxel(x,yy,z,v);
if v.Used then
First := false;
end;

if y = maxy then
First := true;

Second := True;

if y > 0 then
for YY := y-1 downto 0 do
if Second then
begin
voxel.GetVoxel(x,yy,z,v);
if v.Used then
Second := false;
end;

if y = 0 then
Second := true;

Result := false;

If (First) and (not Second) then
begin
B2 := true;
B1 := false;
Result := true;
end
else
If (not First) and (Second) then
begin
B2 := false;
B1 := true;
Result := true;
end;

end;

function CheckFrontBack(var b1,b2 : boolean) : boolean;
var
XX : integer;
First,Second : boolean;
v : TVoxelunpacked;
begin

First := True;

if x < maxx then
for XX := x+1 to Maxx do
if First then
begin
voxel.GetVoxel(xx,y,z,v);
if v.Used then
First := false;
end;

if x = maxx then
First := true;

Second := True;

if x > 0 then
for xx := x-1 downto 0 do
if Second then
begin
voxel.GetVoxel(xx,y,z,v);
if v.Used then
Second := false;
end;

if x = 0 then
Second := true;

Result := false;

If (First) and (not Second) then
begin
B2 := false;
B1 := true;
Result := true;
end
else
If (not First) and (Second) then
begin
B2 := true;
B1 := false;
Result := true;
end;

end;
Procedure CalcFacing;
var v: TVoxelUnpacked;
    B : array [0..5] of boolean;
    I : integer;
    Pos : TVector3i;
begin
     // get the voxel in question
     voxel.GetVoxel(x,y,z,v);
     // skip empty ones
     if not v.Used then Exit;

     for i := 0 to 5 do
     B[i] := false;

     if Empty(x+1,y,z) then B[0] := true;
     if Empty(x-1,y,z) then B[1] := true;
     if Empty(x,y-1,z) then B[2] := true;
     if Empty(x,y+1,z) then B[3] := true;
     if Empty(x,y,z-1) then B[4] := true;
     if Empty(x,y,z+1) then B[5] := true;

     // cancel out double-sided facings
     if B[0] and B[1] then if not CheckFrontBack(B[0],B[1]) then if x > trunc(maxx/2) then B[1] := false else B[0] := false;  //Dec(Flags,ChooseFrontOrBack(x,y,z));
     if B[2] and B[3] then if not CheckRightLeft(B[2],B[3]) then if y < trunc(maxy/2) then B[3] := false else B[2] := false;
     if B[4] and B[5] then if not CheckTopBottom(B[4],B[5]) then if z < trunc(maxz/2) then B[5] := false else B[4] := false;

     Pos := SetVectorI(0,0,0);
     if B[0] then
     Pos.X := Pos.X + 1;
     if B[1] then
     Pos.X := Pos.X - 1;
     if B[2] then
     Pos.Y := Pos.Y - 1;
     if B[3] then
     Pos.Y := Pos.Y + 1;
     if B[4] then
     Pos.Z := Pos.Z - 1;
     if B[5] then
     Pos.Z := Pos.Z + 1;

     if (Pos.x = 0) and (Pos.z = 0) and (Pos.y = 0) then
     begin
     inc(Result);

     Red[Red_No].X := X;
     Red[Red_No].Y := Y;
     Red[Red_No].Z := Z;

     inc(Red_No);
     SetLength(Red,Red_No+1);

     end;
end;
begin
     maxx := voxel.Tailer.XSize - 1;
     maxy := voxel.Tailer.YSize - 1;
     maxz := voxel.Tailer.ZSize - 1;

     Result := 0;

     SetLength(Red,0);
     Red_No := 0;
     SetLength(Red,Red_No+1);

     // examine each used voxel, pass one for outside voxels
     for x := 0 to maxx do
         for y := 0 to maxy do
             for z := 0 to maxz do
                 CalcFacing;

     if Red_No > 0 then
     for x := 0 to Red_No-1 do
     begin
     voxel.GetVoxel(Red[x].x,Red[x].y,Red[x].z,v);
     v.Used := false;
     voxel.SetVoxel(Red[x].x,Red[x].y,Red[x].z,v);
     end;
end;

Function GetNormal3f(Vxl : TVoxelSection; N : integer) : TVector3f;
begin

Result := setvector(0,0,0);

if Vxl.Tailer.Unknown = 2 then
Result := TSNormals[n]
else
if Vxl.Tailer.Unknown = 4 then
Result := RA2Normals[n];

end;

function SmoothNormals(var Vxl : TVoxelSection) : TApplyNormalsResult;
var
   x,y,z,maxx,maxy,maxz,t : integer;
   v : TVoxelUnPacked;
   VoxelSmoothData : array of TVoxelSmoothData;
   Data_no : integer;
   function GetVoxel(x,y,z: integer; var v : TVoxelUnpacked): boolean;
   begin
      // check bounds
      Result := false; // outside the voxel is all empty
      if (x < 0) or (x > maxx) then Exit;
      if (y < 0) or (y > maxy) then Exit;
      if (z < 0) or (z > maxz) then Exit;
      // ok, do the real check
      vxl.GetVoxel(x,y,z,v);
      Result := v.Used;
   end;
   Function GetAverage: integer;
   var
      i : integer;
      Normals : TVector3f;
      Count : integer;
      v : TVoxelUnPacked;
   begin
      GetVoxel(x,y,z,v);
      Count := 0;//1;

      For i := 0 to 25 do
         if GetVoxel(x+trunc(CubeNormals[i].x),y+trunc(CubeNormals[i].y),z+trunc(CubeNormals[i].z),v) then
         begin
            Normals.X := Normals.X + GetNormal3f(Vxl,v.Normal).x;
            Normals.Y := Normals.Y + GetNormal3f(Vxl,v.Normal).y;
            Normals.Z := Normals.Z + GetNormal3f(Vxl,v.Normal).z;
            inc(count);
         end;

      if count > 0 then
      begin
         Normalize(Normals);

         GetVoxel(x,y,z,v);
         Normals.X := Normals.X + GetNormal3f(Vxl,v.Normal).x;
         Normals.Y := Normals.Y + GetNormal3f(Vxl,v.Normal).y;
         Normals.Z := Normals.Z + GetNormal3f(Vxl,v.Normal).z;

         Normalize(Normals);
         Result := Norm2Index(Normals,Vxl);
      end
      else
      begin
         GetVoxel(x,y,z,v);
         Result := v.Normal;
      end;
   end;
begin
   maxx := vxl.Tailer.XSize - 1;
   maxy := vxl.Tailer.YSize - 1;
   maxz := vxl.Tailer.ZSize - 1;

   Result.applied := 0;
   Result.confused := 0;

   Data_no := 0;
   SetLength(VoxelSmoothData,Data_no);

   for x := 0 to maxx do
      for y := 0 to maxy do
         for z := 0 to maxz do
         begin
            vxl.GetVoxel(x,y,z,v);
            if V.Used then
            begin
               T := GetAverage;

               if T = V.Normal then
                  inc(Result.confused)
               else
                  inc(Result.applied);
               V.Normal := T;

               inc(Data_no);
               SetLength(VoxelSmoothData,Data_no);
               VoxelSmoothData[Data_no-1].Pos.X := x;
               VoxelSmoothData[Data_no-1].Pos.Y := y;
               VoxelSmoothData[Data_no-1].Pos.Z := z;
               VoxelSmoothData[Data_no-1].v.Colour := v.Colour;
               VoxelSmoothData[Data_no-1].v.Normal := v.Normal;
               VoxelSmoothData[Data_no-1].v.Flags := v.Flags;
               VoxelSmoothData[Data_no-1].v.Used := v.Used;
               vxl.SetVoxel(x,y,z,v);
            end;
         end;

   SetLength(VoxelSmoothData,0);
end;

// Old Stuff below, used for RemoveRedundentVoxelsOld

const isInside = 0;
      isOutside = 1;

function BitSet(value, mask: integer): boolean;
begin
     Result := (value and mask) = mask;
end;

function FindInside(voxel: TVoxelSection): Integer;
{ This function starts by assuming every cell is inside.
  It then does a flood fill from every face of the cube,
  flood-filling all empty cells to outside.
}
var maxx, maxy, maxz: integer;
procedure ClearFlags;
var v: TVoxelUnpacked;
    x, y, z: integer;
begin
     for x := 0 to maxx do
         for y := 0 to maxy do
             for z := 0 to maxz do begin
                 voxel.GetVoxel(x,y,z,v);
                 v.Flags := isInside;
                 voxel.SetVoxel(x,y,z,v);
             end;
end;
var OutsideCount: integer;
function IsOnOutside(x,y,z: Integer): Boolean;
procedure DoCheck(x,y,z: Integer);
var v: TVoxelUnpacked;
begin
     // bounds check!
     if (x < 0) or (x > maxx) then Exit;
     if (y < 0) or (y > maxy) then Exit;
     if (z < 0) or (z > maxz) then Exit;
     // get the voxel
     voxel.GetVoxel(x,y,z,v);
     // is it outside already?
     if BitSet(v.Flags,isOutside) then
        Result := True;
end;
begin
     Result := False; // assume not
     DoCheck(x+1,y,z);
     DoCheck(x,y+1,z);
     DoCheck(x,y,z+1);
     DoCheck(x-1,y,z);
     DoCheck(x,y-1,z);
     DoCheck(x,y,z-1);
end;
var x, y, z, {Pass,} PassCount: integer;
    v: TVoxelUnpacked;
begin
     maxx := voxel.Tailer.XSize - 1;
     maxy := voxel.Tailer.YSize - 1;
     maxz := voxel.Tailer.ZSize - 1;
     ClearFlags;
     OutsideCount := 0;
     //Pass := 0;
     repeat
         // Inc(Pass);
          // use pass count for memory too
          PassCount := 0;
          // do a scan of all cells
          for x := 0 to maxx do
              for y := 0 to maxy do
                  for z := 0 to maxz do begin
                      voxel.GetVoxel(x,y,z,v);
                      // obvious checks
                      if v.Used then Continue; // used voxels are 'inside'
                      if BitSet(v.Flags,IsOutside) then Continue; // already found?
                      // check for extremities
                      if (x = 0) or (x = maxx) or
                         (y = 0) or (y = maxy) or
                         (z = 0) or (z = maxz) or
                      // or adjacent to an outside
                         IsOnOutside(x,y,z) then begin
                      // then set it to outside
                         v.Flags := isOutside;
                         voxel.SetVoxel(x,y,z,v);
                         Inc(PassCount);
                      end;
                  end;
          // keep running total
          Inc(OutsideCount,PassCount);
     until (Passcount = 0);
     Result := OutsideCount;
end;

// a voxel is internal if all adjacent voxels (including diagonals) are inside
function Internal(voxel: TVoxelSection; x,y,z: Integer): Boolean;
var maxx, maxy, maxz: Integer;
function IsExternal(x,y,z:Integer):Boolean;
var v: TVoxelUnpacked;
begin
     Result := True; // assume is
     if (x < 0) or (x > maxx) then Exit;
     if (y < 0) or (y > maxy) then Exit;
     if (z < 0) or (z > maxz) then Exit;
     voxel.getVoxel(x,y,z,v);
     Result := BitSet(v.Flags,isOutside);
end;
var x1, y1, z1: Integer;
begin
     Result := False; // assume not
     maxx := voxel.Tailer.XSize - 1;
     maxy := voxel.Tailer.YSize - 1;
     maxz := voxel.Tailer.ZSize - 1;
     for x1 := Pred(x) to Succ(x) do
         for y1 := Pred(y) to Succ(y) do
             for z1 := Pred(z) to Succ(z) do
                 if IsExternal(x1,y1,z1) then
                    Exit;
     Result := True;
end;

// This method works with voxels that don't have holes
function RemoveRedundantVoxelsOld(voxel: TVoxelSection): integer;
var x,y,z,
    maxx, maxy, maxz: Integer;
    v: TVoxelUnpacked;
begin
     Result := 0;
     // first, make sure flags are ok
     FindInside(voxel);
     // constraints
     maxx := voxel.Tailer.XSize - 1;
     maxy := voxel.Tailer.YSize - 1;
     maxz := voxel.Tailer.ZSize - 1;
     // now test it
     for x := 0 to maxx do
         for y := 0 to maxy do
             for z := 0 to maxz do
                 if Internal(voxel,x,y,z) then begin
                    voxel.GetVoxel(x,y,z,v);
                    if v.Used then begin
                       v.Used := False;
                       voxel.SetVoxel(x,y,z,v);
                       Inc(Result);
                    end;
                 end;
end;

function GetSmoothNormal(var Vxl : TVoxelSection; X,Y,Z,Normal : integer) : integer;
var
   t,maxx,maxy,maxz : integer;
   v : TVoxelUnPacked;

   // Get voxel data
   function GetVoxel(x,y,z: integer; var v : TVoxelUnpacked): boolean;
   begin
      // check bounds
      Result := false; // outside the voxel is all empty
      if (x < 0) or (x > maxx) then Exit;
      if (y < 0) or (y > maxy) then Exit;
      if (z < 0) or (z > maxz) then Exit;
      // ok, do the real check
      vxl.GetVoxel(x,y,z,v);
      Result := v.Used;
   end;

   Function GetAverage: integer;
   var
      i : integer;
      Normals : TVector3f;
      Count : integer;
      v : TVoxelUnPacked;
   begin
      GetVoxel(x,y,z,v);
//      Result := V.Normal;
      Count := 0;
      For i := 0 to 25 do
         if GetVoxel(x+trunc(CubeNormals[i].x),y+trunc(CubeNormals[i].y),z+trunc(CubeNormals[i].z),v) then
         begin
            Normals.X := Normals.X + GetNormal3f(Vxl,v.Normal).x;
            Normals.Y := Normals.Y + GetNormal3f(Vxl,v.Normal).y;
            Normals.Z := Normals.Z + GetNormal3f(Vxl,v.Normal).z;
            inc(count);
         end;

      if count > 0 then
      begin
         Normalize(Normals);
         GetVoxel(x,y,z,v);
         Normals.X := Normals.X + GetNormal3f(Vxl,v.Normal).x;
         Normals.Y := Normals.Y + GetNormal3f(Vxl,v.Normal).y;
         Normals.Z := Normals.Z + GetNormal3f(Vxl,v.Normal).z;
         Normalize(Normals);
         Result := Norm2Index(Normals,Vxl);
      end
      else
      begin
         GetVoxel(x,y,z,v);
         Result := v.Normal;
      end;
   end;
//
// Main Smmothing Function Starts HERE
//
begin
   maxx := vxl.Tailer.XSize - 1;
   maxy := vxl.Tailer.YSize - 1;
   maxz := vxl.Tailer.ZSize - 1;

   vxl.GetVoxel(x,y,z,v);
   if V.Used then
   begin
      T := GetAverage;
      Result := T;
   end
   else
      Result := Normal;
end;

end.

